#!/usr/bin/env python3
# process_database.py --> MLEndSND

import argparse
from pathlib import Path

import pandas as pd
from nkululeko.utils import files


def main():
    parser = argparse.ArgumentParser(description='Process database')
    parser.add_argument('--data_dir', type=str, default='MLEndSND_Public')
    parser.add_argument('--output_dir', type=str, default='.')
    args = parser.parse_args()

    data_dir = Path(args.data_dir)
    output_dir = Path(args.output_dir)

    paths = files.find_files(data_dir, ext="wav")
    annots = pd.read_csv(data_dir / "MLEndSND_Audio_Attributes.csv", dtype=str)
    annots.set_index("Public filename", inplace=True)
    spk_info = pd.read_csv(
        data_dir / "MLEndSND_Speakers_Demographics.csv", dtype=str)
    spk_info.set_index("Speaker", inplace=True)
    df = annots.join(spk_info, "Speaker")

    # rename columns
    df.rename(columns={"Numeral": "numeral", "Intonation": "emotion", "Speaker": "speaker", "Language1": "language1", "Language2": "language2", "Nationality": "nationality", "Coordinates": "coordinates"}, inplace=True)
    
    df.index.name = "file"

    # add extension .wav to file
    df.index = df.index + ".wav"
    # allocate speaker > 200 for test
    train_df = df[df["speaker"].astype(int) <= 120]
    test_df = df.drop(train_df.index)

    # save to csv
    df.to_csv(output_dir / "mlendsnd.csv")
    train_df.to_csv(output_dir / "mlendsnd_train.csv")
    test_df.to_csv(output_dir / "mlendsnd_test.csv")

    print(f"Total: {len(df)}, Train: {len(train_df)}, Test: {len(test_df)}")

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# process_database.py --> URDU

import argparse
from pathlib import Path

import pandas as pd

emotion_map = {
    "A": "anger",
    "S": "sadness",
    "H": "happiness",
    "N": "neutral",
}


gender_map = {
    "M": "male",
    "F": "female"
}


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--data_dir', type=str,
                        default='URDU', help='Path to the database file')
    parser.add_argument('--output_dir', type=str, default='.', help='Path to the output directory')
    args = parser.parse_args()

    data_dir = Path(args.data_dir)
    output_dir = Path(args.output_dir)

    if not output_dir.exists():
        output_dir.mkdir()

    paths = list(data_dir.glob('**/*.wav'))
    emotions = [emotion_map[p.stem[p.stem.rfind("_") + 1]] for p in paths]
    speakers = [p.stem[: p.stem.index("_")] for p in paths]
    genders = [gender_map[s[1]] for s in speakers]
    languages = ["urdu" for p in paths]

    df = pd.DataFrame({"file": paths, "emotion": emotions, "speaker": speakers, "gender": genders, "language": languages})


    # allocate speaker independent train/test --> very low result
    # df_test = df[df["speaker"].isin(["SM3", "SM5", "SF10", "SF9", "SF11"])]
    # df_train = df.drop(df_test.index)

    # use train_test_split
    from sklearn.model_selection import train_test_split
    df_train, df_test = train_test_split(
        df, test_size=0.2, random_state=42, stratify=df["emotion"])


    # save to csv
    df.to_csv(output_dir / "urdu.csv", index=False)
    df_train.to_csv(output_dir / "urdu_train.csv", index=False)
    df_test.to_csv(output_dir / "urdu_test.csv", index=False)

    print(f"Total: {len(df)}, Train: {len(df_train)}, Test: {len(df_test)}")


if __name__ == "__main__":
    main()



# feats_audmodel.py
import os

import audeer
import audinterface
import audonnx
import nkululeko.glob_conf as glob_conf
import numpy as np
import pandas as pd
import torch
from nkululeko.feat_extract.featureset import Featureset


class AudModelSet(Featureset):
    """
    Embeddings from the wav2vec2. based model finetuned on MSPPodcast emotions, described in the paper
    "Dawn of the transformer era in speech emotion recognition: closing the valence gap"
    https://arxiv.org/abs/2203.07378
    """

    def __init__(self, name, data_df):
        super().__init__(name, data_df)
        model_url = "https://zenodo.org/record/6221127/files/w2v2-L-robust-12.6bc4a7fd-1.1.0.zip"
        model_root = self.util.config_val("FEATS", "aud.model", "./audmodel/")
        if not os.path.isdir(model_root):
            cache_root = audeer.mkdir("cache")
            model_root = audeer.mkdir(model_root)
            archive_path = audeer.download_url(
                model_url, cache_root, verbose=True
            )
            audeer.extract_archive(archive_path, model_root)
        cuda = "cuda" if torch.cuda.is_available() else "cpu"
        device = self.util.config_val("MODEL", "device", cuda)
        self.model = audonnx.load(model_root, device=device)

    def extract(self):
        """Extract the features based on the initialized dataset or re-open them when found on disk."""
        store = self.util.get_path("store")
        store_format = self.util.config_val("FEATS", "store_format", "pkl")
        storage = f"{store}{self.name}.{store_format}"
        extract = eval(
            self.util.config_val("FEATS", "needs_feature_extraction", "False")
        )
        no_reuse = eval(self.util.config_val("FEATS", "no_reuse", "False"))
        if no_reuse or extract or not os.path.isfile(storage):
            self.util.debug(
                "extracting audmodel embeddings, this might take a while..."
            )
            hidden_states = audinterface.Feature(
                self.model.labels("hidden_states"),
                process_func=self.model,
                process_func_args={
                    "outputs": "hidden_states",
                },
                sampling_rate=16000,
                resample=True,
                num_workers=5,
                verbose=True,
            )
            self.df = hidden_states.process_index(self.data_df.index)
            self.util.write_store(self.df, storage, store_format)
            try:
                glob_conf.config["DATA"]["needs_feature_extraction"] = "False"
            except KeyError:
                pass
        else:
            self.util.debug("reusing extracted audmodel features.")
            self.df = self.util.get_store(storage, store_format)

    def extract_sample(self, signal, sr):
        result = self.model(signal, sr)
        return np.asarray(result["hidden_states"].flatten())



# augmenter.py
import os

import audeer
import audiofile
import numpy as np
import pandas as pd
from audformat.utils import map_file_path
from audiomentations import (AddGaussianNoise, AddGaussianSNR, Compose,
                             PitchShift, Shift, TimeStretch)
from nkululeko.util import Util
from tqdm import tqdm


class Augmenter:
    """
    augmenting the train split
    """

    def __init__(self, df):
        self.df = df
        self.util = Util("augmenter")
        # Define a standard transformation that randomly add augmentations to files
        self.audioment = Compose(
            [
                AddGaussianNoise(
                    min_amplitude=0.001, max_amplitude=0.015, p=0.5
                ),
                # AddGaussianSNR(min_snr_db=10, max_snr_db=40, p=0.5),
                TimeStretch(min_rate=0.8, max_rate=1.25, p=0.5),
                PitchShift(min_semitones=-4, max_semitones=4, p=0.5),
                Shift(min_fraction=-0.5, max_fraction=0.5, p=0.5),
            ]
        )

    def changepath(self, fp, np):
        #        parent = os.path.dirname(fp).split('/')[-1]
        fullpath = os.path.dirname(fp)
        #       newpath = f'{np}{parent}'
        #       audeer.mkdir(newpath)
        return fp.replace(fullpath, np)

    def augment(self, sample_selection):
        """
        augment the training files and return a dataframe with new files index.
        """
        files = self.df.index.get_level_values(0).values
        store = self.util.get_path("store")
        filepath = f"{store}augmentations/"
        audeer.mkdir(filepath)
        self.util.debug(f"augmenting {sample_selection} samples to {filepath}")
        newpath = ""
        for i, f in enumerate(tqdm(files)):
            signal, sr = audiofile.read(f)
            filename = os.path.basename(f)
            parent = os.path.dirname(f).split("/")[-1]
            sig_aug = self.audioment(samples=signal, sample_rate=sr)
            newpath = f"{filepath}/{parent}/"
            audeer.mkdir(newpath)
            audiofile.write(
                f"{newpath}{filename}", signal=sig_aug, sampling_rate=sr
            )
        df_ret = self.df.copy()
        df_ret = df_ret.set_index(
            map_file_path(df_ret.index, lambda x: self.changepath(x, newpath))
        )
        aug_db_filename = self.util.config_val(
            "DATA", "augment_result", "augment.csv"
        )
        target = self.util.config_val("DATA", "target", "emotion")
        df_ret[target] = df_ret["class_label"]
        df_ret = df_ret.drop(columns=["class_label"])
        df_ret.to_csv(aug_db_filename)
        return df_ret



list# demo.py
# Demonstration code to use the ML-experiment framework
# Test the loading of a previously trained model and demo mode
# needs the project config file to run before

from nkululeko.experiment import Experiment
import configparser
from nkululeko.util import Util
from  nkululeko.constants import VERSION
import argparse
import os

def main(src_dir):
    parser = argparse.ArgumentParser(description='Call the nkululeko framework.')
    parser.add_argument('--config', default='exp.ini', help='The base configuration')
    parser.add_argument('--file', help='A file that should be processed (16kHz mono wav)')
    parser.add_argument('--list', help='A file with a list of files, one per line, that should be processed (16kHz mono wav)')
    args = parser.parse_args()
    if args.config is not None:
        config_file = args.config    
    else:
        config_file = f'{src_dir}/exp.ini'


    # test if the configuration file exists
    if not os.path.isfile(config_file):
        print(f'ERROR: no such file: {config_file}')
        exit()
        
    # load one configuration per experiment
    config = configparser.ConfigParser()
    config.read(config_file)
    
    # create a new experiment
    expr = Experiment(config)
    util = Util('demo')
    util.debug(f'running {expr.name}, nkululeko version {VERSION}')

    # load the experiment
    expr.load(f'{util.get_save_name()}')
    if args.file is None and args.list is None:
       expr.demo(None, False)
    else:
        if args.list is None:
            expr.demo(args.file, False)
        else:
            expr.demo(args.list, True)

    print('DONE')


if __name__ == "__main__":
    cwd = os.path.dirname(os.path.abspath(__file__))
    main(cwd) # use this if you want to state the config file path on command line


import os

src_path = './demos/multiple_exeriments/'


features = [
    {'--feat': '[\\\'os\\\']'},
    {'--feat': '[\\\'wav2vec\\\']'},
    {'--feat': '[\\\'trill\\\']'},
    {'--feat': '[\\\'praat\\\']'},
    {'--feat': '[\\\'os\\\',\\\'praat\\\']'},
    {'--feat': '[\\\'wav2vec\\\',\\\'praat\\\']'},
    {'--feat': '[\\\'trill\\\',\\\'praat\\\']'},
]


for f in features:
    cmd = f'python {src_path}parse_nkulu.py --config "./tests/exp_multiple_cross.ini" --model mlp --learning_rate 0.001 --drop 0.2 '
    for item in f:
        cmd += f'{item} {f[item]} '
    print(cmd)
    os.system(cmd)


# feats_clap.py

from nkululeko.util import Util
from nkululeko.feat_extract.featureset import Featureset
import os
import pandas as pd
from tqdm import tqdm
import os
import nkululeko.glob_conf as glob_conf
import laion_clap
import audiofile


class Clap(Featureset):
    """Class to extract laion's clap embeddings (https://github.com/LAION-AI/CLAP)"""

    def __init__(self, name, data_df):
        """Constructor. is_train is needed to distinguish from test/dev sets, because they use the codebook from the training"""
        super().__init__(name, data_df)
        self.device = self.util.config_val("MODEL", "device", "cpu")
        self.model_initialized = False

    def init_model(self):
        # load model
        self.util.debug("loading clap model...")
        self.model = laion_clap.CLAP_Module(enable_fusion=False)
        self.model.load_ckpt()  # download the default pretrained checkpoint.
        print(f"loaded clap model")

    def extract(self):
        """Extract the features or load them from disk if present."""
        store = self.util.get_path("store")
        store_format = self.util.config_val("FEATS", "store_format", "pkl")
        storage = f"{store}{self.name}.{store_format}"
        extract = self.util.config_val(
            "FEATS", "needs_feature_extraction", False
        )
        no_reuse = eval(self.util.config_val("FEATS", "no_reuse", "False"))
        if extract or no_reuse or not os.path.isfile(storage):
            if not self.model_initialized:
                self.init_model()
            self.util.debug(
                "extracting clap embeddings, this might take a while..."
            )
            emb_series = pd.Series(index=self.data_df.index, dtype=object)
            length = len(self.data_df.index)
            for idx, (file, start, end) in enumerate(
                tqdm(self.data_df.index.to_list())
            ):
                signal, sampling_rate = audiofile.read(
                    file,
                    offset=start.total_seconds(),
                    duration=(end - start).total_seconds(),
                    always_2d=True,
                )
                emb = self.get_embeddings(signal, sampling_rate)
                emb_series[idx] = emb
            self.df = pd.DataFrame(
                emb_series.values.tolist(), index=self.data_df.index
            )
            self.util.write_store(self.df, storage, store_format)
            try:
                glob_conf.config["DATA"]["needs_feature_extraction"] = "false"
            except KeyError:
                pass
        else:
            self.util.debug("reusing extracted wav2vec2 embeddings")
            self.df = self.util.get_store(storage, store_format)
            if self.df.isnull().values.any():
                nanrows = self.df.columns[self.df.isna().any()].tolist()
                print(nanrows)
                self.util.error(
                    f"got nan: {self.df.shape} {self.df.isnull().sum().sum()}"
                )

    def get_embeddings(self, signal, sampling_rate):
        tmp_audio_name = ["clap_audio_tmp.wav"]
        audiofile.write(tmp_audio_name[0], signal, 48000)
        audio_embed = self.model.get_audio_embedding_from_filelist(
            x=tmp_audio_name, use_tensor=False
        )
        return audio_embed[0]

    def extract_sample(self, signal, sr):
        self.init_model()
        feats = self.get_embeddings(signal, sr)
        return feats



#!/home/audeering.local/fburkhardt/research/nkululeko/venv/bin/python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2014 CNRS

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# AUTHORS
# Hervé BREDIN - http://herve.niderb.fr

"""
Hidden Markov Model with (constrained) Viterbi decoding

Usage:
  hmm train [-g <gaussian>] [-c <covariance>] [-d <duration>] <uris.lst> <references.mdtm> <features.pkl> <model.pkl>
  hmm apply [-d <duration>] [-f <constraint.mdtm>] <model.pkl> <features.pkl> <hypothesis.mdtm>
  hmm -h | --help
  hmm --version

Options:
  -g <gaussian>            Number of gaussian components [default: 16].
  -c <covariance>          Covariance type (diag or full) [default: diag].
  -d <duration>            Minimum duration in seconds [default: 0.250].
  -f <constraint.mdtm>     Constrain Viterbi decoding to follow this path.
  -h --help                Show this screen.
  --version                Show version.
"""

from pyannote.algorithms.segmentation.hmm import ViterbiHMM
from pyannote.parser.util import CoParser
from pyannote.parser import MDTMParser
from docopt import docopt
import pickle


def do_train(
    uris_lst, references_mdtm, features_pkl, model_pkl,
    n_components=16, covariance_type='diag', min_duration=0.250,
):

    hmm = ViterbiHMM(
        n_components=n_components, covariance_type=covariance_type,
        random_state=None, thresh=1e-2, min_covar=1e-3, n_iter=10,
        disturb=0.05, sampling=1000, min_duration=min_duration)

    # iterate over all uris in a synchronous manner
    coParser = CoParser(uris=uris_lst,
                        reference=references_mdtm,
                        features=features_pkl)
    references, features = coParser.generators('reference', 'features')

    hmm.fit(references, features)

    with open(model_pkl, 'wb') as f:
        pickle.dump(hmm, f)


def do_apply(model_pkl, features_pkl, hypothesis_mdtm,
             min_duration=0.250, constraint_mdtm=None):

    with open(model_pkl, 'rb') as f:
        hmm = pickle.load(f)

    hmm.min_duration = min_duration

    with open(features_pkl, 'rb') as f:
        features = pickle.load(f)

    constraint = None
    if constraint_mdtm:
        constraint = MDTMParser().read(constraint_mdtm)()

    hypothesis = hmm.apply(features, constraint=constraint)

    with open(hypothesis_mdtm, 'w') as f:
        MDTMParser().write(hypothesis, f=f)


if __name__ == '__main__':

    arguments = docopt(__doc__, version='Hidden Markov Models 1.0')

    if arguments['train']:

        uris_lst = arguments['<uris.lst>']
        references_mdtm = arguments['<references.mdtm>']
        features_pkl = arguments['<features.pkl>']
        model_pkl = arguments['<model.pkl>']

        n_components = int(arguments['-g'])
        covariance_type = arguments['-c']
        min_duration = float(arguments['-d'])

        do_train(
            uris_lst, references_mdtm, features_pkl, model_pkl,
            n_components=n_components, covariance_type=covariance_type,
            min_duration=min_duration)

    elif arguments['apply']:

        model_pkl = arguments['<model.pkl>']
        features_pkl = arguments['<features.pkl>']
        hypothesis_mdtm = arguments['<hypothesis.mdtm>']
        min_duration = float(arguments['-d'])
        constraint_mdtm = arguments['-f']

        do_apply(model_pkl, features_pkl, hypothesis_mdtm,
                 min_duration=min_duration, constraint_mdtm=constraint_mdtm)



"""
Code originally by Oliver Pauly

Based on an idea by Klaus Scherer

K. R. Scherer, “Randomized splicing: A note on a simple technique for masking speech content” 
Journal of Experimental Research in Personality, vol. 5, pp. 155–159, 1971.

Evaluated in:
F. Burkhardt, Anna Derington, Matthias Kahlau, Klaus Scherer, Florian Eyben and Björn Schuller: Masking Speech Contents by Random Splicing: is Emotional Expression Preserved?, Proc. ICASSP, 2023

"""
import numpy as np
import librosa
import audiofile as af


def random_splicing(
    signal,
    sr,
    p_reverse=0.0,
    top_db=12,
):
    """
    randomly splice the signal and re-arrange.

    p_reverse: probability of some samples to be in reverse order
    top_db:  top db level for silence to be recognized

    """
    signal /= np.max(abs(signal))

    indices = split_wav_naive(signal, top_db=top_db)

    np.random.shuffle(indices)

    wav_spliced = remix_random_reverse(signal, indices, p_reverse=p_reverse)

    return wav_spliced


def split_wav_naive(wav, top_db=12):
    indices = librosa.effects.split(wav, top_db=top_db)

    indices = np.array(indices)
    # (re)add the silence-segments
    indices = indices.repeat(2)[1:-1].reshape((-1, 2))
    # add first segment
    indices = np.vstack(((0, indices[0][0]), indices))
    # add last segment
    indices = np.vstack((indices, [indices[-1][-1], wav.shape[0]]))

    return indices


def remix_random_reverse(wav, indices, p_reverse=0):
    import random

    wav_remix = []

    for seg in indices:
        start = seg[0]
        end = seg[1]
        wav_seg = wav[start:end]

        if np.random.rand(1)[0] <= p_reverse:
            wav_seg = wav_seg[::-1]

        wav_remix.append(wav_seg)

    wav_remix = np.hstack(wav_remix)

    return wav_remix



# main.py
# Demonstration code to use the ML-experiment framework

import sys
sys.path.append("./nkululeko")
import experiment as exp
import configparser
from util import Util
import constants
        
def main(config_file):
    # load one configuration per experiment
    config = configparser.ConfigParser()
    config.read(config_file)
    
    # create a new experiment
    expr = exp.Experiment(config)

    util = Util()

    util.debug(f'running {expr.name}, nkululeko version {constants.VERSION}')

    # load the data
    expr.load_datasets()

    # split into train and test
    expr.fill_train_and_tests()
    util.debug(f'train shape : {expr.df_train.shape}, test shape:{expr.df_test.shape}')

    # extract features
    expr.extract_feats()
#    util.debug(f'train feats shape : {expr.feats_train.df.shape}, test feats shape:{expr.feats_test.df.shape}')

    # initialize a run manager
    expr.init_runmanager()

    # run the experiment
    expr.run()

    print('DONE')

if __name__ == "__main__":

    main('./tests/exp_irony_class.ini')


# segment.py
# segment data splits

import argparse
import os
import pandas as pd
import configparser
from nkululeko.experiment import Experiment
from nkululeko.util import Util
from nkululeko.constants import VERSION
import nkululeko.glob_conf as glob_conf
from nkululeko.reporting.report_item import ReportItem

def main(src_dir):
    parser = argparse.ArgumentParser(
        description="Call the nkululeko framework."
    )
    parser.add_argument(
        "--config", default="exp.ini", help="The base configuration"
    )
    args = parser.parse_args()
    if args.config is not None:
        config_file = args.config
    else:
        config_file = f"{src_dir}/exp.ini"

    # test if the configuration file exists
    if not os.path.isfile(config_file):
        print(f"ERROR: no such file: {config_file}")
        exit()

    # load one configuration per experiment
    config = configparser.ConfigParser()
    config.read(config_file)
    # create a new experiment
    expr = Experiment(config)
    util = Util("segment")
    util.debug(f"running {expr.name}, nkululeko version {VERSION}")

    if util.config_val("EXP", "no_warnings", False):
        import warnings

        warnings.filterwarnings("ignore")

    # load the data
    expr.load_datasets()

    # split into train and test
    expr.fill_train_and_tests()
    util.debug(
        f"train shape : {expr.df_train.shape}, test shape:{expr.df_test.shape}"
    )

    # segment
    segment_target = util.config_val("SEGMENT", "target", "_seg")

    segmenter = util.config_val("SEGMENT", "method", "silero")
    sample_selection = util.config_val("SEGMENT", "sample_selection", "all")
    if sample_selection == "all":
        df = pd.concat([expr.df_train, expr.df_test])
    elif sample_selection == "train":
        df = expr.df_train
    elif sample_selection == "test":
        df = expr.df_test
    else:
        util.error(
            f"unknown segmentation selection specifier {sample_selection},"
            " should be [all | train | test]"
        )
    util.debug(
        f"segmenting {sample_selection}: {df.shape[0]} samples with {segmenter}"
    )
    if segmenter == "silero":
        from nkululeko.segmenting.seg_silero import Silero_segmenter

        segmenter = Silero_segmenter()
        df_seg = segmenter.segment_dataframe(df)

    else:
        util.error(f"unkown segmenter: {segmenter}")

    def calc_dur(x):
        from datetime import datetime

        starts = x[1]
        ends = x[2]
        return (ends - starts).total_seconds()

    df_seg["duration"] = df_seg.index.to_series().map(lambda x: calc_dur(x))
    if "duration" not in df.columns:
        df["duration"] = df.index.to_series().map(lambda x: calc_dur(x))
    num_before = df.shape[0]
    num_after = df_seg.shape[0]
    # plot distributions
    from nkululeko.plots import Plots

    plots = Plots()
    plots.plot_durations(df, "original_durations", sample_selection, caption='Original durations')
    plots.plot_durations(df_seg, "segmented_durations", sample_selection, caption='Segmented durations')
    print("")
    # remove encoded labels
    target = util.config_val("DATA", "target", "emotion")
    if "class_label" in df_seg.columns:
        df_seg = df_seg.drop(columns=[target])
        df_seg = df_seg.rename(columns={"class_label": target})
    # save file
    dataname = "_".join(expr.datasets.keys())
    name = f"{dataname}{segment_target}"
    df_seg.to_csv(f"{expr.data_dir}/{name}.csv")
    util.debug(
        f"saved {name}.csv to {expr.data_dir}, {num_after} samples (was"
        f" {num_before})"
    )
    glob_conf.report.add_item(
        ReportItem(
            "Data",
            'Segmentation',
            f'Segmented {num_before} samples to {num_after} segments',
        )
    )
    expr.store_report()
    print("DONE")


def get_segmentation(file):
    #    print(f'segmenting {file[0]}')
    print(".", end="")
    wav = read_audio(file[0], sampling_rate=SAMPLING_RATE)
    speech_timestamps = get_speech_timestamps(
        wav, vad_model, sampling_rate=SAMPLING_RATE
    )
    files, starts, ends = [], [], []
    for entry in speech_timestamps:
        start = float(entry["start"] / 1000.0)
        end = float(entry["end"] / 1000.0)
        files.append(file[0])
        starts.append(start)
        ends.append(end)
    seg_index = segmented_index(files, starts, ends)
    return seg_index


def segment_dataframe(df):
    dfs = []
    for file, values in df.iterrows():
        index = get_segmentation(file)
        dfs.append(
            pd.DataFrame(
                values.to_dict(),
                index,
            )
        )
    return audformat.utils.concat(dfs)


if __name__ == "__main__":
    cwd = os.path.dirname(os.path.abspath(__file__))
    main(
        cwd
    )  # use this if you want to state the config file path on command line



# cnnmodel.py

from sklearn.utils import resample
from nkululeko.models.model import Model
import torch
from sklearn.metrics import mean_squared_error
import nkululeko.glob_conf as glob_conf
from nkululeko.reporter import Reporter
import numpy as np
from nkululeko.losses.loss_ccc import ConcordanceCorCoeff


class CNN_model(Model):
    """A CNN (convolutional neural net) model"""

    is_classifier = True

    def __init__(self, df_train, df_test, feats_train, feats_test):
        """Constructor taking the configuration and all dataframes"""
        Model.__init__(self, df_train, df_test, feats_train, feats_test)
        self.util.debug(f"initializing model")
        self.device = glob_conf.config["MODEL"]["device"]
        pretrained_state = self.util.config_val(
            "MODEL", "pre_train", "gender_state.pth.tar"
        )
        state = torch.load(pretrained_state)
        state.pop("out.gender.weight")
        state.pop("out.gender.bias")
        state["fc1.weight"] = state.pop("fc1.gender.weight")
        state["fc1.bias"] = state.pop("fc1.gender.bias")
        model = audpann.Cnn10(sampling_rate=16000, output_dim=1)
        model.load_state_dict(state, strict=False)
        self.model = model.to(self.device)
        self.loss_func = self.util.config_val("MODEL", "loss", "mse")
        if self.loss_func == "mse":
            self.criterion = torch.nn.MSELoss()
        elif self.loss_func == "1-ccc":
            self.criterion = ConcordanceCorCoeff()
        else:
            self.util.error(f"unknown loss function: {self.loss_func}")
        self.util.debug(f"training model with {self.loss_func} loss function")
        self.learning_rate = float(
            self.util.config_val("MODEL", "learning_rate", 0.0001)
        )
        self.optimizer = torch.optim.Adam(
            self.model.parameters(), lr=self.learning_rate
        )

    def train(self):
        """Train the model one epoch"""
        losses = []
        # first check if the model already has been trained
        # if os.path.isfile(self.store_path):
        #     self.load(self.run, self.epoch)
        #     self.util.debug(f'reusing model: {self.store_path}')
        #     return

        self.model.train()

        for features, labels in self.feats_train:
            logits = self.model(features.to(self.device).float()).squeeze(1)
            loss = self.criterion(logits, labels.float().to(self.device))
            losses.append(loss.item())
            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()
        self.loss = (np.asarray(losses)).mean()

    def predict(self):
        """Predict the whole eval feature set"""
        # evaluate on dev set
        _, truths, predictions = self.evaluate_model(False)
        # evaluate on train set, if there is one
        result = 0
        if self.feats_train != None:
            result, _, _ = self.evaluate_model(True)
        report = Reporter(
            truths.numpy(), predictions.numpy(), self.run, self.epoch
        )
        try:
            report.result.loss = self.loss
        except (
            AttributeError
        ):  # if the model was loaded from disk the loss is unknown
            pass
        report.result.train = result
        return report

    def evaluate_model(self, on_train=False):
        if on_train:
            loader = self.feats_train
        else:
            loader = self.feats_test
        logits = torch.zeros(len(loader.dataset))
        targets = torch.zeros(len(loader.dataset))
        self.model.eval()
        with torch.no_grad():
            for index, (features, labels) in enumerate(loader):
                start_index = index * loader.batch_size
                end_index = (index + 1) * loader.batch_size
                if end_index > len(loader.dataset):
                    end_index = len(loader.dataset)
                logits[start_index:end_index] = self.model(
                    features.to(self.device).float()
                ).squeeze(1)
                targets[start_index:end_index] = labels

        measure = self.util.config_val("MODEL", "measure", "mse")
        if measure == "mse":
            result = mean_squared_error(targets.numpy(), logits.numpy())
        elif measure == "ccc":
            result = Reporter.ccc(targets.numpy(), logits.numpy())
        else:
            self.util.error(f"unknown measure: {measure}")
        return result, targets, logits

    def store(self):
        torch.save(self.model.state_dict(), self.store_path)
        self.device = self.util.config_val("MODEL", "device", "cpu")
        # self.model.to(self.device)

    def load(self, run, epoch):
        self.set_id(run, epoch)
        dir = self.util.get_path("model_dir")
        name = (
            f"{self.util.get_exp_name(only_train=True)}_{run}_{epoch:03d}.model"
        )
        self.device = self.util.config_val("MODEL", "device", "cpu")
        self.store_path = dir + name
        self.model = audpann.Cnn10(sampling_rate=16000, output_dim=1)
        state_dict = torch.load(dir + name, map_location="cpu")
        self.model.load_state_dict(state_dict, strict=False)
        self.model.to(self.device)



# loss_concordance_cor_coeff.py

import torch


class ConcordanceCorCoeff(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.mean = torch.mean
        self.var = torch.var
        self.sum = torch.sum
        self.sqrt = torch.sqrt
        self.std = torch.std

    def forward(self, prediction, ground_truth):
        mean_gt = self.mean(ground_truth, 0)
        mean_pred = self.mean(prediction, 0)
        var_gt = self.var(ground_truth, 0)
        var_pred = self.var(prediction, 0)
        v_pred = prediction - mean_pred
        v_gt = ground_truth - mean_gt
        cor = self.sum(v_pred * v_gt) / (
            self.sqrt(self.sum(v_pred**2)) * self.sqrt(self.sum(v_gt**2))
        )
        sd_gt = self.std(ground_truth)
        sd_pred = self.std(prediction)
        numerator = 2 * cor * sd_gt * sd_pred
        denominator = var_gt + var_pred + (mean_gt - mean_pred) ** 2
        ccc = numerator / denominator
        return 1 - ccc



# process_database.py --> CaFE
# modified from https://github.com/Strong-AI-Lab/emotion/blob/master/datasets/CaFE/process.py

import argparse
from pathlib import Path

import pandas as pd

emotion_map = {
    "C": "anger",
    "D": "disgust",
    "J": "happiness",
    "N": "neutral",
    "P": "fear",
    "S": "surprise",
    "T": "sadness",
}

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("--data_dir", type=str, default="CaFE")
    parser.add_argument("--output_dir", type=str, default=".")
    args = parser.parse_args()
    
    data_dir = Path(args.data_dir)
    output_dir = Path(args.output_dir)

    paths = list(data_dir.glob("**/*.wav"))
    files = [file for file in paths]
    emotion = [emotion_map[file.stem[3]] for file in files]
    speaker = [file.stem[:2] for file in files]
    language = ["french" for file in files]
    country = ["canada" for file in files]
    gender = [["female", "male"][int(v) % 2] for v in speaker]
    
    # convert to df
    df = pd.DataFrame({"file": files, "speaker": speaker, "emotion": emotion, "language": language, "country": country})

    # allocate the last two speakers (11, 12) for test
    train_df = df[(df["speaker"] != "11") & (df["speaker"] != "12")]
    test_df = df.drop(train_df.index)

    # save to CSV
    df.to_csv(output_dir / "cafe.csv", index=False)
    train_df.to_csv(output_dir / "cafe_train.csv", index=False)
    test_df.to_csv(output_dir / "cafe_test.csv", index=False)


if __name__ == "__main__":
    main()



# model_knn_reg.py

from sklearn.neighbors import KNeighborsRegressor
from nkululeko.models.model import Model


class KNN_reg_model(Model):
    """An KNN model"""

    is_classifier = False

    def __init__(self, df_train, df_test, feats_train, feats_test):
        super().__init__(df_train, df_test, feats_train, feats_test)
        method = self.util.config_val("MODEL", "KNN_weights", "uniform")
        k = int(self.util.config_val("MODEL", "K_val", "5"))
        self.clf = KNeighborsRegressor(
            n_neighbors=k, weights=method
        )  # set up the classifier



#!/usr/bin/env python3
# process_database.py --> MSED

import argparse
from email import parser
from pathlib import Path

import pandas as pd
from sklearn.model_selection import train_test_split

gender_map = {
    "F": "female",
    "M": "male",
    "C": "child",
}


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--data_dir', type=str,
                        default='MESD/Mexican Emotional Speech Database (MESD)', help='Path to the database file')
    parser.add_argument('--output_dir', type=str, default='.', help='Path to the output directory')
    args = parser.parse_args()

    data_dir = Path(args.data_dir)
    output_dir = Path(args.output_dir)

    if not output_dir.exists():
        output_dir.mkdir()
        
    paths = list(data_dir.glob('**/*.wav'))
    emotions = [p.stem.split("_")[0].lower() for p in paths]
    voices = [p.stem.split("_")[1] for p in paths]
    genders = [gender_map[p.stem.split("_")[1]] for p in paths]
    languages = ["spanish" for p in paths]
    corpus = [p.stem.split("_")[2] for p in paths]
    
    if str(data_dir).find("reduced") != -1:
        reduced_level = [p.stem.split("_")[3] for p in paths]
        words = [" ".join(p.stem.split("_")[4:]) for p in paths]
        df = pd.DataFrame({"file": paths, "emotion": emotions, "gender": genders, "language": languages, "reduced_level": reduced_level, "word": words})       

    else:
        words = [" ".join(p.stem.split("_")[3:]) for p in paths]
        df = pd.DataFrame({"file": paths, "emotion": emotions, "gender": genders, "language": languages, "word": words})

    # split train and test
    df_train, df_test = train_test_split(df, test_size=0.2, random_state=42, stratify=df["emotion"])
    
    # save to csv
    df.to_csv(output_dir / "mesd.csv", index=False)
    df_train.to_csv(output_dir / "mesd_train.csv", index=False)
    df_test.to_csv(output_dir / "mesd_test.csv", index=False)

    print(f"Total: {len(df)}, Train: {len(df_train)}, Test: {len(df_test)}")


if __name__ == "__main__":
    main()



# demo_model.py
# Demonstration code to use the ML-experiment framework
# Test the loading of a previously trained model and demo mode
# needs the project tests/exp_emodb to run before

import sys
sys.path.append("./src")
from nkululeko.experiment import Experiment
import configparser
from nkululeko.util import Util
from  nkululeko.constants import VERSION

def main(config_file):
    # load one configuration per experiment
    config = configparser.ConfigParser()
    config.read(config_file)
    
    # create a new experiment
    expr = Experiment(config)
    util = Util('demo')
    util.debug(f'running {expr.name}, nkululeko version {VERSION}')

    # load the experiment
    expr.load(f'{util.get_save_name()}')

    expr.demo()

    print('DONE')


if __name__ == "__main__":
    main('tests/exp_emodb.ini')



# process_database.py --> ekorpus
# modified from https://github.com/Strong-AI-Lab/emotion/blob/master/datasets/EESC/process.py
# need to install textgrid: pip install textgrid

import argparse
from pathlib import Path

import pandas as pd
import textgrid
from sklearn.model_selection import train_test_split
from tqdm import tqdm


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--data_dir', type=str, default='EKORPUS')
    parser.add_argument('--output_dir', type=str, default='.')
    args = parser.parse_args()

    data_dir = Path(args.data_dir)
    output_dir = Path(args.output_dir)

    paths = list(data_dir.glob('*.wav'))

    labels = {}
    sentences = {}
    for path in tqdm(
        data_dir.glob("*.TextGrid"), desc="Processing annotations", total=len(paths)
    ):
        grid = textgrid.TextGrid.fromFile(path)
        labels[path.stem] = grid.getFirst("emotion")[0].mark
        sentences[path.stem] = grid.getFirst("sentence")[0].mark

    emotion = labels.values()
    language = ['estonian' for _ in range(len(paths))]

    df = pd.DataFrame(
        {"file": paths, "emotion": emotion, "language": language})

    # split for training and test
    df_train, df_test = train_test_split(
        df, test_size=0.2, random_state=42, stratify=df['emotion'])


    # save to CSV
    df.to_csv(output_dir / 'ekorpus.csv', index=False)
    df_train.to_csv(output_dir / 'ekorpus_train.csv', index=False)
    df_test.to_csv(output_dir / 'ekorpus_test.csv', index=False)

    print(f"EKORPUS: {len(df)} samples, {len(df_train)} train, {len(df_test)} test")


if __name__ == '__main__':
    main()



# randomsplicer.py

"""
Code originally by Oliver Pauly

Based on an idea by Klaus Scherer

K. R. Scherer, “Randomized splicing: A note on a simple technique for masking speech content” 
Journal of Experimental Research in Personality, vol. 5, pp. 155–159, 1971.

Evaluated in:
F. Burkhardt, Anna Derington, Matthias Kahlau, Klaus Scherer, Florian Eyben and Björn Schuller: Masking Speech Contents by Random Splicing: is Emotional Expression Preserved?, Proc. ICASSP, 2023

"""

import pandas as pd
from tqdm import tqdm
import os
import numpy as np
import audiofile as af
from audformat.utils import map_file_path
import audeer
from nkululeko.util import Util
import nkululeko.augmenting.randomsplicing as rsp


class Randomsplicer:
    """
    augmenting the train split
    """

    def __init__(self, df):
        self.df = df
        self.util = Util("randomsplicer")

    def changepath(self, fp, np):
        fullpath = os.path.dirname(fp)
        return fp.replace(fullpath, np)

    def run(self, sample_selection):
        """
        random splice the selected samples and return a dataframe with new files index.
            adjustable parameters:
            * p_reverse: probability of some samples to be in reverse order (default: 0.3)
            * top_db: top db level for silence to be recognized (default: 12)
        """

        p_reverse = 0.3
        top_db = 12

        files = self.df.index.get_level_values(0).values
        store = self.util.get_path("store")
        filepath = f"{store}randomspliced/"
        audeer.mkdir(filepath)
        self.util.debug(
            f"random splicing {sample_selection} samples to {filepath}"
        )
        newpath = ""
        for i, f in enumerate(tqdm(files)):
            signal, sr = af.read(f)
            filename = os.path.basename(f)
            parent = os.path.dirname(f).split("/")[-1]
            sig_new = rsp.random_splicing(
                signal,
                sr,
                p_reverse=p_reverse,
                top_db=top_db,
            )

            newpath = f"{filepath}/{parent}/"
            audeer.mkdir(newpath)
            af.write(f"{newpath}{filename}", signal=sig_new, sampling_rate=sr)
        df_ret = self.df.copy()
        df_ret = df_ret.set_index(
            map_file_path(df_ret.index, lambda x: self.changepath(x, newpath))
        )
        db_filename = self.util.config_val(
            "DATA", "random_splice_result", "random_spliced.csv"
        )
        target = self.util.config_val("DATA", "target", "emotion")
        df_ret[target] = df_ret["class_label"]
        df_ret = df_ret.drop(columns=["class_label"])
        df_ret.to_csv(db_filename)
        return df_ret



import ast
import os
import pickle
import random
import time
from cgi import print_environ  # To convert strings to objects

import audeer
import audformat
import numpy as np
import pandas as pd
from sklearn.preprocessing import LabelEncoder

import nkululeko.glob_conf as glob_conf
from nkululeko.data.dataset import Dataset
from nkululeko.data.dataset_csv import Dataset_CSV
from nkululeko.demo_predictor import Demo_predictor
from nkululeko.feat_extract.feats_analyser import FeatureAnalyser
from nkululeko.feature_extractor import FeatureExtractor
from nkululeko.file_checker import FileChecker
from nkululeko.filter_data import DataFilter, filter_min_dur
from nkululeko.plots import Plots
from nkululeko.reporting.report import Report
from nkululeko.runmanager import Runmanager
from nkululeko.scaler import Scaler
from nkululeko.test_predictor import Test_predictor
from nkululeko.util import Util


class Experiment:
    """Main class specifying an experiment"""

    def __init__(self, config_obj):
        """
        Parameters
        ----------
        config_obj : a config parser object that sets the experiment parameters and being set as a global object.
        """

        self.set_globals(config_obj)
        self.name = glob_conf.config["EXP"]["name"]
        self.root = glob_conf.config["EXP"]["root"]
        self.data_dir = self.root + self.name
        audeer.mkdir(self.data_dir)  # create the experiment directory
        self.util = Util("experiment")
        glob_conf.set_util(self.util)
        try:
            with open(self.data_dir + "/report.pkl", "rb") as handle:
                self.report = pickle.load(handle)
        except FileNotFoundError:
            self.report = Report()
        glob_conf.set_report(self.report)
        self.loso = self.util.config_val("MODEL", "loso", False)
        self.logo = self.util.config_val("MODEL", "logo", False)
        self.xfoldx = self.util.config_val("MODEL", "k_fold_cross", False)
        self.start = time.process_time()

    def store_report(self):
        with open(self.data_dir + "/report.pkl", "wb") as handle:
            pickle.dump(self.report, handle)
        if eval(self.util.config_val("REPORT", "show", "False")):
            self.report.print()
        if self.util.config_val("REPORT", "latex", False):
            self.report.export_latex() 

    def get_name(self):
        return self.util.get_exp_name()

    def set_globals(self, config_obj):
        """install a config object in the global space"""
        glob_conf.init_config(config_obj)

    def load_datasets(self):
        """Load all databases specified in the configuration and map the labels"""
        ds = ast.literal_eval(glob_conf.config["DATA"]["databases"])
        self.datasets = {}
        self.got_speaker, self.got_gender, self.got_age = False, False, False
        for d in ds:
            ds_type = self.util.config_val_data(d, "type", "audformat")
            if ds_type == "audformat":
                data = Dataset(d)
            elif ds_type == "csv":
                data = Dataset_CSV(d)
            else:
                self.util.error(f"unknown data type: {ds_type}")
            data.load()
            data.prepare()
            if data.got_gender:
                self.got_gender = True
            if data.got_age:
                self.got_age = True
            if data.got_speaker:
                self.got_speaker = True
            self.datasets.update({d: data})
        self.target = self.util.config_val("DATA", "target", "emotion")
        # print target via debug
        self.util.debug(f"target: {self.target}")
        # print keys/column
        dbs = ",".join(list(self.datasets.keys()))
        labels = self.util.config_val("DATA", "labels", False)
        if labels:
            labels = ast.literal_eval(labels)
        else:
            labels = list(
                next(iter(self.datasets.values())).df[self.target].unique()
            )
        # print labels via debug
        self.util.debug(f"Target labels (user defined): {labels}")
        glob_conf.set_labels(labels)
        self.util.debug(f"loaded databases {dbs}")

    def _import_csv(self, storage):
        # df = pd.read_csv(storage, header=0, index_col=[0,1,2])
        # df.index.set_levels(pd.to_timedelta(df.index.levels[1]), level=1)
        # df.index.set_levels(pd.to_timedelta(df.index.levels[2]), level=2)
        df = audformat.utils.read_csv(storage)
        df.is_labeled = True if self.target in df else False
        # print(df.head())
        return df

    def fill_tests(self):
        """Only fill a new test set"""

        test_dbs = ast.literal_eval(glob_conf.config["DATA"]["tests"])
        self.df_test = pd.DataFrame()
        start_fresh = eval(self.util.config_val("DATA", "no_reuse", "False"))
        store = self.util.get_path("store")
        storage_test = f"{store}extra_testdf.csv"
        if os.path.isfile(storage_test) and not start_fresh:
            self.util.debug(f"reusing previously stored {storage_test}")
            self.df_test = self._import_csv(storage_test)
        else:
            for d in test_dbs:
                ds_type = self.util.config_val_data(d, "type", "audformat")
                if ds_type == "audformat":
                    data = Dataset(d)
                elif ds_type == "csv":
                    data = Dataset_CSV(d)
                else:
                    self.util.error(f"unknown data type: {ds_type}")
                data.load()
                if data.got_gender:
                    self.got_gender = True
                if data.got_age:
                    self.got_age = True
                if data.got_speaker:
                    self.got_speaker = True
                data.prepare_labels()
                self.df_test = pd.concat(
                    [self.df_test, self.util.make_segmented_index(data.df)]
                )
                self.df_test.is_labeled = data.is_labeled
            self.df_test.got_gender = self.got_gender
            self.df_test.got_speaker = self.got_speaker
            # self.util.set_config_val('FEATS', 'needs_features_extraction', 'True')
            # self.util.set_config_val('FEATS', 'no_reuse', 'True')
            self.df_test["class_labels"] = self.df_test[self.target]
            self.df_test[self.target] = self.label_encoder.transform(
                self.df_test[self.target]
            )
            self.df_test.to_csv(storage_test)

    def fill_train_and_tests(self):
        """Set up train and development sets. The method should be specified in the config."""
        store = self.util.get_path("store")
        storage_test = f"{store}testdf.csv"
        storage_train = f"{store}traindf.csv"
        start_fresh = eval(self.util.config_val("DATA", "no_reuse", "False"))
        if (
            os.path.isfile(storage_train)
            and os.path.isfile(storage_test)
            and not start_fresh
        ):
            self.util.debug(
                f"reusing previously stored {storage_test} and {storage_train}"
            )
            self.df_test = self._import_csv(storage_test)
            # print(f"df_test: {self.df_test}")
            self.df_train = self._import_csv(storage_train)
            # print(f"df_train: {self.df_train}")
        else:
            self.df_train, self.df_test = pd.DataFrame(), pd.DataFrame()
            for d in self.datasets.values():
                d.split()
                d.prepare_labels()
                if d.df_train.shape[0] == 0:
                    self.util.debug(f"warn: {d.name} train empty")
                self.df_train = pd.concat([self.df_train, d.df_train])
                # print(f"df_train: {self.df_train}")
                self.util.copy_flags(d, self.df_train)
                if d.df_test.shape[0] == 0:
                    self.util.debug(f"warn: {d.name} test empty")
                self.df_test = pd.concat([self.df_test, d.df_test])
                self.util.copy_flags(d, self.df_test)
            store = self.util.get_path("store")
            storage_test = f"{store}testdf.csv"
            storage_train = f"{store}traindf.csv"
            self.df_test.to_csv(storage_test)
            self.df_train.to_csv(storage_train)

        self.util.copy_flags(self, self.df_test)
        self.util.copy_flags(self, self.df_train)
        # Try data checks
        datachecker = FileChecker(self.df_train)
        self.df_train = datachecker.all_checks()
        datachecker.set_data(self.df_test)
        self.df_test = datachecker.all_checks()

        # Check for filters
        filter_sample_selection = self.util.config_val(
            "DATA", "filter.sample_selection", "all"
        )
        if filter_sample_selection == "all":
            datafilter = DataFilter(self.df_train)
            self.df_train = datafilter.all_filters()
            datafilter = DataFilter(self.df_test)
            self.df_test = datafilter.all_filters()
        elif filter_sample_selection == "train":
            datafilter = DataFilter(self.df_train)
            self.df_train = datafilter.all_filters()
        elif filter_sample_selection == "test":
            datafilter = DataFilter(self.df_test)
            self.df_test = datafilter.all_filters()
        else:
            self.util.error(
                "unkown filter sample selection specifier"
                f" {filter_sample_selection}, should be [all | train | test]"
            )

        # encode the labels
        if self.util.exp_is_classification():
            datatype = self.util.config_val("DATA", "type", "dummy")
            if datatype == "continuous":
                if self.df_test.is_labeled:
                    # remember the target in case they get labelencoded later
                    self.df_test["class_label"] = self.df_test[self.target]
                    test_cats = self.df_test["class_label"].unique().astype(
                        'str')
                else:
                    # if there is no target, copy a dummy label
                    self.df_test = self._add_random_target(self.df_test)
                if self.df_train.is_labeled:
                    # remember the target in case they get labelencoded later
                    self.df_train["class_label"] = self.df_train[self.target]
                    train_cats = self.df_train["class_label"].unique().astype(
                        'str')
            else:
                if self.df_test.is_labeled:
                    test_cats = self.df_test[self.target].unique().astype(
                        'str')
                else:
                    # if there is no target, copy a dummy label
                    self.df_test = self._add_random_target(
                        self.df_test).astype('str')
                train_cats = self.df_train[self.target].unique().astype('str')
                # print(f"df_train: {pd.DataFrame(self.df_train[self.target])}")
                # print(f"train_cats with target {self.target}: {train_cats}")
            if self.df_test.is_labeled:
                if type(test_cats) == np.ndarray:
                    self.util.debug(f"Categories test (nd.array): {test_cats}")
                else:
                    self.util.debug(f"Categories test (list): {test_cats.to_list()}")
            if type(train_cats) == np.ndarray:
                self.util.debug(f"Categories train (nd.array): {train_cats}")
            else:
                self.util.debug(f"Categories train (list): {train_cats.to_list()}")

            # encode the labels as numbers
            self.label_encoder = LabelEncoder()
            self.df_train[self.target] = self.label_encoder.fit_transform(
                self.df_train[self.target]
            )
            self.df_test[self.target] = self.label_encoder.transform(
                self.df_test[self.target]
            )
            glob_conf.set_label_encoder(self.label_encoder)
        if self.got_speaker:
            self.util.debug(
                f"{self.df_test.speaker.nunique()} speakers in test and"
                f" {self.df_train.speaker.nunique()} speakers in train"
            )

        target_factor = self.util.config_val("DATA", "target_divide_by", False)
        if target_factor:
            self.df_test[self.target] = self.df_test[self.target] / float(
                target_factor
            )
            self.df_train[self.target] = self.df_train[self.target] / float(
                target_factor
            )
            if not self.util.exp_is_classification():
                self.df_test["class_label"] = self.df_test[
                    "class_label"
                ] / float(target_factor)
                self.df_train["class_label"] = self.df_train[
                    "class_label"
                ] / float(target_factor)

    def _add_random_target(self, df):
        labels = glob_conf.labels
        a = [None] * len(df)
        for i in range(0, len(df)):
            a[i] = random.choice(labels)
        df[self.target] = a
        return df

    def plot_distribution(self):
        """Plot the distribution of samples and speaker per target class and biological sex"""
        plot = Plots()
        sample_selection = self.util.config_val(
            "EXPL", "sample_selection", "all"
        )
        if sample_selection == "all":
            df_labels = pd.concat([self.df_train, self.df_test])
            self.util.copy_flags(self.df_train, df_labels)
        elif sample_selection == "train":
            df_labels = self.df_train
            self.util.copy_flags(self.df_train, df_labels)
        elif sample_selection == "test":
            df_labels = self.df_test
            self.util.copy_flags(self.df_test, df_labels)
        else:
            self.util.error(
                f"unkown sample selection specifier {sample_selection}, should"
                " be [all | train | test]"
            )

        plot.plot_distributions(df_labels)
        if self.got_speaker:
            plot.plot_distributions_speaker(df_labels)

    def extract_test_feats(self):
        self.feats_test = pd.DataFrame()
        feats_name = "_".join(
            ast.literal_eval(glob_conf.config["DATA"]["tests"])
        )
        feats_types = self.util.config_val_list("FEATS", "type", ["os"])
        self.feature_extractor = FeatureExtractor(
            self.df_test, feats_types, feats_name, "test"
        )
        self.feats_test = self.feature_extractor.extract()
        self.util.debug(f"Test features shape:{self.feats_test.shape}")

    def extract_feats(self):
        """Extract the features for train and dev sets.

        They will be stored on disk and need to be removed manually.

        The string FEATS.feats_type is read from the config, defaults to os.

        """
        df_train, df_test = self.df_train, self.df_test
        feats_name = "_".join(
            ast.literal_eval(glob_conf.config["DATA"]["databases"])
        )
        self.feats_test, self.feats_train = pd.DataFrame(), pd.DataFrame()
        feats_types = self.util.config_val_list("FEATS", "type", ["os"])
        self.feature_extractor = FeatureExtractor(
            df_train, feats_types, feats_name, "train"
        )
        self.feats_train = self.feature_extractor.extract()
        self.feature_extractor = FeatureExtractor(
            df_test, feats_types, feats_name, "test"
        )
        self.feats_test = self.feature_extractor.extract()
        self.util.debug(
            f"All features: train shape : {self.feats_train.shape}, test"
            f" shape:{self.feats_test.shape}"
        )
        if self.feats_train.shape[0] < self.df_train.shape[0]:
            self.util.warn(
                f"train feats ({self.feats_train.shape[0]}) != train labels"
                f" ({self.df_train.shape[0]})"
            )
            self.df_train = self.df_train[
                self.df_train.index.isin(self.feats_train.index)
            ]
            self.util.warn(f"mew train labels shape: {self.df_train.shape[0]}")
        if self.feats_test.shape[0] < self.df_test.shape[0]:
            self.util.warn(
                f"test feats ({self.feats_test.shape[0]}) != test labels"
                f" ({self.df_test.shape[0]})"
            )
            self.df_test = self.df_test[
                self.df_test.index.isin(self.feats_test.index)
            ]
            self.util.warn(f"mew test labels shape: {self.df_test.shape[0]}")

        self._check_scale()

    def augment(self):
        """
        Augment the selected samples
        """
        from nkululeko.augmenting.augmenter import Augmenter

        sample_selection = self.util.config_val("DATA", "augment", "train")
        if sample_selection == "all":
            df = pd.concat([self.df_train, self.df_test])
        elif sample_selection == "train":
            df = self.df_train
        elif sample_selection == "test":
            df = self.df_test
        else:
            self.util.error(
                f"unknown augmentation selection specifier {sample_selection},"
                " should be [all | train | test]"
            )

        augmenter = Augmenter(df)
        augmenter.augment(sample_selection)

    def autopredict(self):
        """
        Predict labels for samples with existing models and add to the dataframe.
        """
        sample_selection = self.util.config_val("PREDICT", "split", "all")
        if sample_selection == "all":
            df = pd.concat([self.df_train, self.df_test])
        elif sample_selection == "train":
            df = self.df_train
        elif sample_selection == "test":
            df = self.df_test
        else:
            self.util.error(
                f"unknown augmentation selection specifier {sample_selection},"
                " should be [all | train | test]"
            )
        targets = self.util.config_val_list("PREDICT", "targets", ["gender"])
        for target in targets:
            if target == "gender":
                from nkululeko.autopredict.ap_gender import GenderPredictor

                predictor = GenderPredictor(df)
                df = predictor.predict(sample_selection)
            elif target == "age":
                from nkululeko.autopredict.ap_age import AgePredictor

                predictor = AgePredictor(df)
                df = predictor.predict(sample_selection)
            elif target == "snr":
                from nkululeko.autopredict.ap_snr import SNRPredictor

                predictor = SNRPredictor(df)
                df = predictor.predict(sample_selection)
            elif target == "mos":
                from nkululeko.autopredict.ap_mos import MOSPredictor

                predictor = MOSPredictor(df)
                df = predictor.predict(sample_selection)
            elif target == "pesq":
                from nkululeko.autopredict.ap_pesq import PESQPredictor

                predictor = PESQPredictor(df)
                df = predictor.predict(sample_selection)
            elif target == "sdr":
                from nkululeko.autopredict.ap_sdr import SDRPredictor

                predictor = SDRPredictor(df)
                df = predictor.predict(sample_selection)
            elif target == "stoi":
                from nkululeko.autopredict.ap_stoi import STOIPredictor

                predictor = STOIPredictor(df)
                df = predictor.predict(sample_selection)
            elif target == "arousal":
                from nkululeko.autopredict.ap_arousal import ArousalPredictor

                predictor = ArousalPredictor(df)
                df = predictor.predict(sample_selection)
            elif target == "valence":
                from nkululeko.autopredict.ap_valence import ValencePredictor

                predictor = ValencePredictor(df)
                df = predictor.predict(sample_selection)
            elif target == "dominance":
                from nkululeko.autopredict.ap_dominance import \
                    DominancePredictor

                predictor = DominancePredictor(df)
                df = predictor.predict(sample_selection)
            else:
                self.util.error(f"unknown auto predict target: {target}")
        return df

    def random_splice(self):
        """
        Random-splice the selected samples
        """
        from nkululeko.augmenting.randomsplicer import Randomsplicer

        sample_selection = self.util.config_val(
            "DATA", "random_splice", "train"
        )
        if sample_selection == "all":
            df = pd.concat([self.df_train, self.df_test])
        elif sample_selection == "train":
            df = self.df_train
        elif sample_selection == "test":
            df = self.df_test
        else:
            self.util.error(
                f"unknown augmentation selection specifier {sample_selection},"
                " should be [all | train | test]"
            )
        randomsplicer = Randomsplicer(df)
        randomsplicer.run(sample_selection)

    def analyse_features(self, needs_feats):
        """
        Do a feature exploration

        """

        plot_feats = eval(
            self.util.config_val("EXPL", "feature_distributions", "False")
        )
        sample_selection = self.util.config_val(
            "EXPL", "sample_selection", "all"
        )

        if self.util.config_val("EXPL", "value_counts", False):
            self.plot_distribution()
        if not needs_feats:
            return
        if sample_selection == "all":
            df_feats = pd.concat([self.feats_train, self.feats_test])
            df_labels = pd.concat([self.df_train, self.df_test])
        elif sample_selection == "train":
            df_feats = self.feats_train
            df_labels = self.df_train
        elif sample_selection == "test":
            df_feats = self.feats_test
            df_labels = self.df_test
        else:
            self.util.error(
                f"unknown sample selection specifier {sample_selection}, should"
                " be [all | train | test]"
            )

        if plot_feats:
            feat_analyser = FeatureAnalyser(
                sample_selection, df_labels, df_feats
            )
            feat_analyser.analyse()

        # check if a scatterplot should be done
        scatter_var = eval(self.util.config_val("EXPL", "scatter", "False"))
        if scatter_var:
            scatters = ast.literal_eval(glob_conf.config["EXPL"]["scatter"])
            if self.util.exp_is_classification():
                plots = Plots()
                for scatter in scatters:
                    plots.scatter_plot(
                        df_feats, df_labels["class_label"], scatter
                    )
            else:
                self.util.debug("can't do scatterplot if not classification")

        # check if a tsne should be plotted
        tsne = eval(self.util.config_val("EXPL", "tsne", "False"))
        if tsne:
            if self.util.exp_is_classification():
                plots = Plots()
                all_feats = self.feats_train.append(self.feats_test)
                all_labels = self.df_train["class_label"].append(
                    self.df_test["class_label"]
                )
                plots.plotTsne(
                    all_feats, all_labels, self.util.get_exp_name() + "_tsne"
                )
            else:
                self.util.debug("can't plot tsne if not classification")

    def _check_scale(self):
        scale = self.util.config_val("FEATS", "scale", False)
        if scale:
            self.scaler = Scaler(
                self.df_train,
                self.df_test,
                self.feats_train,
                self.feats_test,
                scale,
            )
            self.feats_train, self.feats_test = self.scaler.scale()

    def init_runmanager(self):
        """Initialize the manager object for the runs."""
        self.runmgr = Runmanager(
            self.df_train, self.df_test, self.feats_train, self.feats_test
        )

    def run(self):
        """Do the runs."""
        self.runmgr.do_runs()

        # access the best results all runs
        self.reports = self.runmgr.best_results
        # try to save yourself
        save = self.util.config_val("EXP", "save", False)
        if save:
            # save the experiment for future use
            self.save(self.util.get_save_name())

        # self.__collect_reports()
        self.util.print_best_results(self.reports)

        # check if the test predictions should be saved to disk
        test_pred_file = self.util.config_val("EXP", "save_test", False)
        if test_pred_file:
            self.predict_test_and_save(test_pred_file)

        # check if the majority voting for all speakers should be plotted
        conf_mat_per_speaker_function = self.util.config_val(
            "PLOT", "combine_per_speaker", False
        )
        if conf_mat_per_speaker_function:
            self.plot_confmat_per_speaker(conf_mat_per_speaker_function)
        used_time = time.process_time() - self.start
        self.util.debug(f"Done, used {used_time:.3f} seconds")

        # check if a test set should be labeled by the model:
        label_data = self.util.config_val("DATA", "label_data", False)
        label_result = self.util.config_val("DATA", "label_result", False)
        if label_data and label_result:
            self.predict_test_and_save(label_result)

        return self.reports

    def plot_confmat_per_speaker(self, function):
        if self.loso or self.logo or self.xfoldx:
            self.util.debug(
                "plot combined speaker predictions not possible for cross"
                " validation"
            )
            return
        best = self._get_best_report(self.reports)
        # if not best.is_classification:
        #     best.continuous_to_categorical()
        truths = best.truths
        preds = best.preds
        speakers = self.df_test.speaker.values
        print(f"{len(truths)} {len(preds)} {len(speakers) }")
        df = pd.DataFrame(
            data={"truth": truths, "pred": preds, "speaker": speakers}
        )
        plot_name = "result_combined_per_speaker"
        self.util.debug(
            f"plotting speaker combination ({function}) confusion matrix to"
            f" {plot_name}"
        )
        best.plot_per_speaker(df, plot_name, function)

    def _get_best_report(self, best_reports):
        return self.runmgr.get_best_result(best_reports)

    def print_best_model(self):
        self.runmgr.print_best_result_runs()

    def demo(self, file, is_list):
        model = self.runmgr.get_best_model()
        labelEncoder = None
        try:
            labelEncoder = self.label_encoder
        except AttributeError:
            pass
        demo = Demo_predictor(
            model, file, is_list, self.feature_extractor, labelEncoder
        )
        demo.run_demo()

    def predict_test_and_save(self, result_name):
        model = self.runmgr.get_best_model()
        model.set_testdata(self.df_test, self.feats_test)
        test_predictor = Test_predictor(
            model, self.df_test, self.label_encoder, result_name
        )
        test_predictor.predict_and_store()

    def load(self, filename):
        f = open(filename, "rb")
        tmp_dict = pickle.load(f)
        f.close()
        self.__dict__.update(tmp_dict)

    def save(self, filename):
        try:
            f = open(filename, "wb")
            pickle.dump(self.__dict__, f)
            f.close()
        except (AttributeError, TypeError, RuntimeError) as error:
            self.util.warn(
                f"Save experiment: Can't pickle local object: {error}"
            )



# feats_oxbow.py

from nkululeko.util import Util
from nkululeko.feat_extract.featureset import Featureset
import os
import pandas as pd
import opensmile


class Openxbow(Featureset):
    """Class to extract openXBOW processed opensmile features (https://github.com/openXBOW)"""

    def __init__(self, name, data_df, is_train=False):
        """Constructor. is_train is needed to distinguish from test/dev sets, because they use the codebook from the training"""
        super().__init__(name, data_df)
        self.is_train = is_train

    def extract(self):
        """Extract the features or load them from disk if present."""
        self.featset = self.util.config_val("FEATS", "set", "eGeMAPSv02")
        self.feature_set = eval(f"opensmile.FeatureSet.{self.featset}")
        store = self.util.get_path("store")
        storage = f"{store}{self.name}_{self.featset}.pkl"
        extract = self.util.config_val(
            "FEATS", "needs_feature_extraction", False
        )
        no_reuse = eval(self.util.config_val("FEATS", "no_reuse", "False"))
        if extract or no_reuse or not os.path.isfile(storage):
            # extract smile features first
            self.util.debug(
                "extracting openSmile features, this might take a while..."
            )
            smile = opensmile.Smile(
                feature_set=self.feature_set,
                feature_level=opensmile.FeatureLevel.LowLevelDescriptors,
                num_workers=5,
            )
            if isinstance(self.data_df.index, pd.MultiIndex):
                is_multi_index = True
                smile_df = smile.process_index(self.data_df.index)
            else:
                smile_df = smile.process_files(self.data_df.index)
            smile_df.index = smile_df.index.droplevel(1)
            smile_df.index = smile_df.index.droplevel(1)
            # compute xbow features
            # set some file names on disk
            lld_name, xbow_name, codebook_name = (
                "llds.csv",
                "xbow.csv",
                "xbow_codebook",
            )
            # save the smile features
            smile_df.to_csv(lld_name, sep=";", header=False)
            # get the path of the xbow java jar file
            xbow_path = self.util.config_val(
                "FEATS", "xbow.model", "../openXBOW/"
            )
            # get the size of the codebook
            size = self.util.config_val("FEATS", "size", 500)
            # get the number of assignements
            assignments = self.util.config_val("FEATS", "assignments", 10)
            # differentiate between train and test
            if self.is_train:
                # store the codebook
                os.system(
                    f"java -jar {xbow_path}openXBOW.jar -i"
                    f" {lld_name} -standardizeInput -log                     -o"
                    f" {xbow_name} -size {size} -a {assignments} -B"
                    f" {codebook_name}"
                )
            else:
                # use the codebook
                os.system(
                    f"java -jar {xbow_path}openXBOW.jar -i {lld_name}          "
                    f"           -o {xbow_name} -b {codebook_name}"
                )
            # read in the result from disk
            xbow_df = pd.read_csv(xbow_name, sep=";", header=None)
            # set the index
            xbow_df = xbow_df.set_index(self.data_df.index)
            # check if smile features should be added
            with_os = self.util.config_val("FEATS", "with_os", False)
            if with_os:
                # extract smile functionals
                self.util.debug(
                    "extracting openSmile functionals, this might take a"
                    " while..."
                )
                smile = opensmile.Smile(
                    feature_set=opensmile.FeatureSet.eGeMAPSv02,  # always use eGemaps for this
                    feature_level=opensmile.FeatureLevel.Functionals,
                    num_workers=5,
                )
                if isinstance(self.data_df.index, pd.MultiIndex):
                    is_multi_index = True
                    smile_df = smile.process_index(self.data_df.index)
                else:
                    smile_df = smile.process_files(self.data_df.index)
                # drop the multi index
                smile_df.index = smile_df.index.droplevel(1)
                smile_df.index = smile_df.index.droplevel(1)
                xbow_df = xbow_df.join(smile_df)
            # in any case, store to disk for later use
            xbow_df.to_pickle(storage)
            # and assign to be the "official" feature set
            self.df = xbow_df
        else:
            self.util.debug("reusing extracted OS features.")
            self.df = pd.read_pickle(storage)



# explore.py
# explore the feature sets

from nkululeko.experiment import Experiment
import configparser
from nkululeko.util import Util
from  nkululeko.constants import VERSION
import argparse
import os

def main(src_dir):
    parser = argparse.ArgumentParser(description='Call the nkululeko framework.')
    parser.add_argument('--config', default='exp.ini', help='The base configuration')
    args = parser.parse_args()
    if args.config is not None:
        config_file = args.config    
    else:
        config_file = f'{src_dir}/exp.ini'

    # test if the configuration file exists
    if not os.path.isfile(config_file):
        print(f'ERROR: no such file: {config_file}')
        exit()
        
    # load one configuration per experiment
    config = configparser.ConfigParser()
    config.read(config_file)
    # create a new experiment
    expr = Experiment(config)
    util = Util()
    util.debug(f'running {expr.name}, nkululeko version {VERSION}')

    # load the data
    expr.load_datasets()

    # split into train and test
    expr.fill_train_and_tests()
    util.debug(f'train shape : {expr.df_train.shape}, test shape:{expr.df_test.shape}')

    # extract features
    expr.extract_feats()

    # explore
    expr.analyse_features()

    print('DONE')


if __name__ == "__main__":
    cwd = os.path.dirname(os.path.abspath(__file__))
    main(cwd) # use this if you want to state the config file path on command line


# process_database.py -> CMU-MOSEI database
# input file:
# - label_paths: CMU_MOSEI_Labels.csv
# - db_paths: CMU-MOSEI

import os
import pandas as pd
import argparse

def preprocess(data_dirs, name, split_df, out_path):
    meta_data = []
    for ix, row in split_df.iterrows():
        # get wav file path
        # print(row['file'])
        filename = row['file'] + '_' + str(row['index']) + '.wav'
        file = os.path.join(data_dirs, 'Raw', 'Audio', 'Segmented_Audio',
                            name, filename)
        
        sentiment = str(row['label2a'])
        for r in (("0", "negative"), ("1", "positive")):
            sentiment = sentiment.replace(*r)
        
        emotion = str(row['label6'])
        for r in (("0", "hap"), ("1", "sad"), ("2", "ang"), ("3", "sur"), ("4", "dis"), ("5", "fea")):
            emotion = emotion.replace(*r)
        
        meta_data.append({
                'file': file,
                'sentiment': sentiment,
                'emotion': emotion,
                })
        
        # write to csv
    meta_data_df = pd.DataFrame(meta_data)
    meta_data_df.to_csv(out_path, index=False)    
    print(f'Wrote {name} partition to {out_path}.')

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--data_dir', type=str, default='./CMU-MOSEI',
                        help='Path to CMU-MOSEI directory.')
    args = parser.parse_args()
    label_path = 'CMU_MOSEI_Labels.csv'
    data = pd.read_csv(label_path)
    for i, split_name in enumerate(['train', 'dev', 'test']):
        print(f'Processing {split_name} (split == {i}).')
        split_df = data[data['split'] == i]
        preprocess(args.data_dir, split_name, split_df, 
                   f'mosei_{split_name}.csv')

    


# predict.py
# use some model and add automatically predicted labels to train and test splits, than save as a new dataset

from nkululeko.experiment import Experiment
import configparser
from nkululeko.util import Util
from nkululeko.constants import VERSION
import argparse
import os


def main(src_dir):
    parser = argparse.ArgumentParser(
        description="Call the nkululeko framework."
    )
    parser.add_argument(
        "--config", default="exp.ini", help="The base configuration"
    )
    args = parser.parse_args()
    if args.config is not None:
        config_file = args.config
    else:
        config_file = f"{src_dir}/exp.ini"

    # test if the configuration file exists
    if not os.path.isfile(config_file):
        print(f"ERROR: no such file: {config_file}")
        exit()

    # load one configuration per experiment
    config = configparser.ConfigParser()
    config.read(config_file)
    # create a new experiment
    expr = Experiment(config)
    util = Util("predict")
    util.debug(
        f"running {expr.name} from config {config_file}, nkululeko version"
        f" {VERSION}"
    )

    # load the data
    expr.load_datasets()

    # split into train and test
    expr.fill_train_and_tests()
    util.debug(
        f"train shape : {expr.df_train.shape}, test shape:{expr.df_test.shape}"
    )

    # process the data
    df = expr.autopredict()
    target = util.config_val("DATA", "target", "emotion")
    if "class_label" in df.columns:
        df = df.drop(columns=[target])
        df = df.rename(columns={"class_label": target})
    name = util.get_data_name() + "_predicted"
    df.to_csv(f"{expr.data_dir}/{name}.csv")
    util.debug(f"saved {name}.csv to {expr.data_dir}")
    print("DONE")


if __name__ == "__main__":
    cwd = os.path.dirname(os.path.abspath(__file__))
    main(
        cwd
    )  # use this if you want to state the config file path on command line



# opensmileset.py
from nkululeko.feat_extract.featureset import Featureset
import opensmile
import os
import pandas as pd
import nkululeko.glob_conf as glob_conf
import ast


class Opensmileset(Featureset):
    def __init__(self, name, data_df):
        super().__init__(name, data_df)
        self.featset = self.util.config_val("FEATS", "set", "eGeMAPSv02")
        try:
            self.feature_set = eval(f"opensmile.FeatureSet.{self.featset}")
            #'eGeMAPSv02, ComParE_2016, GeMAPSv01a, eGeMAPSv01a':
        except AttributeError:
            self.util.error(
                f"something is wrong with feature set: {self.featset}"
            )
        self.featlevel = self.util.config_val("FEATS", "level", "functionals")
        try:
            self.featlevel = self.featlevel.replace(
                "lld", "LowLevelDescriptors"
            )
            self.featlevel = self.featlevel.replace(
                "functionals", "Functionals"
            )
            self.feature_level = eval(
                f"opensmile.FeatureLevel.{self.featlevel}"
            )
        except AttributeError:
            self.util.error(
                f"something is wrong with feature level: {self.featlevel}"
            )

    def extract(self):
        """Extract the features based on the initialized dataset or re-open them when found on disk."""
        store = self.util.get_path("store")
        store_format = self.util.config_val("FEATS", "store_format", "pkl")
        storage = f"{store}{self.name}.{store_format}"
        extract = eval(
            self.util.config_val("FEATS", "needs_feature_extraction", "False")
        )
        no_reuse = eval(self.util.config_val("FEATS", "no_reuse", "False"))
        if extract or not os.path.isfile(storage) or no_reuse:
            self.util.debug(
                "extracting openSmile features, this might take a while..."
            )
            smile = opensmile.Smile(
                feature_set=self.feature_set,
                feature_level=self.feature_level,
                num_workers=5,
                verbose=True,
            )
            if isinstance(self.data_df.index, pd.MultiIndex):
                self.df = smile.process_index(self.data_df.index)
                self.df = self.df.set_index(self.data_df.index)
            else:
                self.df = smile.process_files(self.data_df.index)
                self.df.index = self.df.index.droplevel(1)
                self.df.index = self.df.index.droplevel(1)
            self.util.write_store(self.df, storage, store_format)
            try:
                glob_conf.config["DATA"]["needs_feature_extraction"] = "False"
            except KeyError:
                pass
        else:
            self.util.debug(f"reusing extracted OS features: {storage}.")
            self.df = self.util.get_store(storage, store_format)

    def extract_sample(self, signal, sr):
        smile = opensmile.Smile(
            feature_set=self.feature_set,
            feature_level=opensmile.FeatureLevel.Functionals,
        )
        feats = smile.process_signal(signal, sr)
        return feats.to_numpy()

    def filter(self):
        # use only the features that are indexed in the target dataframes
        self.df = self.df[self.df.index.isin(self.data_df.index)]
        try:
            # use only some features
            selected_features = ast.literal_eval(
                glob_conf.config["FEATS"]["os.features"]
            )
            self.util.debug(
                f"selecting features from opensmile: {selected_features}"
            )
            sel_feats_df = pd.DataFrame()
            hit = False
            for feat in selected_features:
                try:
                    sel_feats_df[feat] = self.df[feat]
                    hit = True
                except KeyError:
                    pass
            if hit:
                self.df = sel_feats_df
                self.util.debug(
                    "new feats shape after selecting opensmile features:"
                    f" {self.df.shape}"
                )
        except KeyError:
            pass



# feats_hubert.py
# Speaker embedding feature extractor for Nkululeko
# https://huggingface.co/speechbrain/spkrec-xvect-voxceleb
# supported feat_type:
# "spkrec-xvect-voxceleb", "spkrec-ecapa-voxceleb", "spkrec-resnet-voxceleb"


import os

import nkululeko.glob_conf as glob_conf
import pandas as pd
import torch
import torchaudio
from nkululeko.feat_extract.featureset import Featureset
from speechbrain.pretrained import EncoderClassifier
from tqdm import tqdm

# from transformers import HubertModel, Wav2Vec2FeatureExtractor


class Spkrec(Featureset):
    """Class to extract SpeechBrain Speaker Embedding embedding)"""

    def __init__(self, name, data_df, feat_type):
        """Constructor. is_train is needed to distinguish from test/dev sets,
        because they use the codebook from the training"""
        super().__init__(name, data_df)
        # check if device is not set, use cuda if available
        cuda = "cuda" if torch.cuda.is_available() else "cpu"
        self.device = self.util.config_val("MODEL", "device", cuda)
        self.classifier_initialized = False
        if feat_type == None:
            self.feat_type = "spkrec-ecapa-voxceleb"
        self.feat_type = feat_type

    def init_model(self):
        # load model
        self.util.debug("loading Spkrec model...")

        model_path = self.util.config_val(
            "FEATS", "Spkrec.model", f"speechbrain/{self.feat_type}"
        )
        self.classifier = EncoderClassifier.from_hparams(model_path)
        print(f"intialized SB model on {self.device}")
        # self.classifier.eval()
        self.classifier_initialized = True

    def extract(self):
        """Extract the features or load them from disk if present."""
        store = self.util.get_path("store")
        storage = f"{store}{self.name}.pkl"
        extract = self.util.config_val(
            "FEATS", "needs_feature_extraction", False
        )
        no_reuse = eval(self.util.config_val("FEATS", "no_reuse", "False"))
        if extract or no_reuse or not os.path.isfile(storage):
            if not self.classifier_initialized:
                self.init_model()
            self.util.debug(
                "extracting Spkrec embeddings, this might take a while..."
            )
            emb_series = pd.Series(index=self.data_df.index, dtype=object)
            length = len(self.data_df.index)
            for idx, (file, start, end) in enumerate(
                tqdm(self.data_df.index.to_list())
            ):
                signal, sampling_rate = torchaudio.load(
                    file,
                    frame_offset=int(start.total_seconds() * 16000),
                    num_frames=int((end - start).total_seconds() * 16000),
                )
                assert sampling_rate == 16000
                # check if signal is stereo, if so, take first channel
                if signal.shape[0] == 2:
                    signal = signal[0]
                emb = self.get_embeddings(signal, sampling_rate, file)
                # fill series with embeddings
                emb_series.iloc[idx] = emb
            # print(f"emb_series shape: {emb_series.shape}")
            self.df = pd.DataFrame(
                emb_series.values.tolist(), index=self.data_df.index
            )
            print(f"df shape: {self.df.shape}")
            self.df.to_pickle(storage)
            try:
                glob_conf.config["DATA"]["needs_feature_extraction"] = "false"
            except KeyError:
                pass
        else:
            self.util.debug(f"reusing extracted {self.feat_type} embeddings")
            self.df = pd.read_pickle(storage)
            if self.df.isnull().values.any():
                nanrows = self.df.columns[self.df.isna().any()].tolist()
                print(nanrows)
                self.util.error(
                    f"got nan: {self.df.shape} {self.df.isnull().sum().sum()}"
                )

    def get_embeddings(self, signal, sampling_rate, file):
        r"""Extract embeddings from raw audio signal."""
        # classifier = EncoderClassifier.from_hparams("speechbrain/spkrec-ecapa-voxceleb")
        try:
            # with torch.no_grad():
            y = self.classifier.encode_batch(signal)
            y = y.squeeze().detach().cpu().numpy()
        except RuntimeError as re:
            print(str(re))
            self.util.error(f"couldn't extract file: {file}")
        # print(f"y ravel shape: {y.ravel().shape}")
        # if y.ravel().shape[0] != 192:
        #     self.util.error(f"got wrong embedding size: {y.ravel().shape} from file: {file}")
        # print(f"y: {y}")
        return y.ravel()

    def extract_sample(self, signal, sr):
        self.init_model()
        feats = self.get_embeddings(signal, sr, "no file")
        return feats
